from meterviewer.datasets import config, dataset
from meterviewer import F


def test_read_single_digit(root_path):
    p = config.read_single_digit_rect(root_path / r"lens_6/XL/XL/M1L1XL" / "config" / "res.xml")

    assert len(p) == 6
    p: list[config.RectO]
    assert p[0].xmax == "105", p[0]


def test_exist_dataset(root_path):
    assert dataset.get_dataset_path(root_path, r"M1L3XL").exists()


def test_read_xml(root_path):
    v, r = config.read_xml(
        root_path / r"lens_6/XL/XL/M1L3XL" / "baocun" / "2018-11-23-12-16-01.xml", config.read_rect_from_node
    )
    assert v == "000994"
    assert r.get("xmin") == "79"


def test_img_to_xml(root_path):
    path_list = [
        config.get_xml_config_path(root_path / r"lens_6/XL/XL/M1L3XL" / "2018-11-23-12-16-01.jpg"),
        config.get_xml_config_path(root_path / r"lens_6/XL/XL/M1L3XL" / "2018-11-23-12-16-01.jpg", "single"),
    ]

    def path_exist(p):
        assert p.exists(), p

    F.must_loop(path_list, path_exist, custom_error=Exception)



import streamlit as st
import pandas as pd
import pathlib
import numpy as np
import os
from meterviewer.config import get_root_path


def our_app():
    st.title("Quick Viewer of Meterdata")
    root_path = get_root_path()
    root_path = st.text_input("Enter the root path", value=str(root_path))
    root_path = pathlib.Path(root_path)

    path = st.text_input("Enter the path of the data", value="generated_merged")
    st.text(os.listdir(root_path / path))

    num = st.text_input("Enter the number of data")
    x_name = st.text_input("Enter the value of x_name", value="x_test.npy")
    y_name = st.text_input("Enter the value of y_name", value="y_test.npy")

    try:
        data_load_state = st.text("reading data...")
        num = int(num)
        data_load_state.text("Done!")
    except ValueError:
        data_load_state.text("Please enter a number")
        return

    folder_path = root_path / path
    x_path = folder_path / x_name
    y_path = folder_path / y_name

    x = np.load(x_path)
    y = np.load(y_path)
    st.text(f"x_shape: {x.shape}, y_shape: {y.shape}")

    st.image(x[num], caption=f"Meterdata {num}")
    st.text(f"Meterdata {num} is {y[num]}")


our_app()



from meterviewer.datasets import imgv
from meterviewer.views import current
from tests.utils import show_img


def test_view_merge_np(root_path):
    """test view_merge_np"""
    d = root_path / "generated"
    current.view_merge_np(d)
    assert True


def test_view_imgv(root_path):
    """test view_imgv"""
    img_path = root_path / "lens_6/XL/XL/M1L3XL/2018-11-23-12-16-01.jpg"
    # assert False, img_path
    im, v, rect = imgv.view_one_img_v(img_path)
    show_img(im)
    assert v == "000994"



from meterviewer.models import littledb
from meterviewer.datasets import dataset
from pathlib import Path as P


def test_insert_one(root_path, remove_files):
    dataset_name = "M1L1XL"
    p = dataset.get_dataset_path(root_path, dataset_name)
    db = p / "items-temp.db"
    insert_one, _ = littledb.create_db(str(db))
    assert db.exists()

    insert_one("test", 0, False)

    # seems not work under windows
    remove_files(db)
    # db.unlink()


def test_create_db(root_path, remove_files):
    p = P("test.db")
    littledb.create_db(str(p))
    remove_files(p)

    p = P(root_path / "test.db")
    littledb.create_db(str(p))
    remove_files(p)



from __future__ import annotations
from meterviewer.models.littledb import get_session, Item
from pathlib import Path as P
from sqlalchemy import ScalarResult, select, text, update


def get_first_item(dbpath: P) -> Item:
    session = get_session(str(dbpath))
    stmt = select(Item).where(Item.id == 1)
    result = session.execute(stmt)
    return result.scalar_one()


def get_carry_items(dbpath: P | str) -> ScalarResult[Item]:
    # get items with is_carry == 1
    session = get_session(dbpath)
    stmt = select(Item).where(Item.is_carry == 1)
    result = session.execute(stmt)
    return result.scalars()


def update_schema(dbpath: P):
    # add column carry_num to Item
    session = get_session(str(dbpath))
    session.execute(text("ALTER TABLE Item ADD COLUMN carry_num INTEGER"))
    session.commit()


def update_item(dbpath: P, id: int, carry_num: int):
    session = get_session(dbpath)
    if carry_num != 0:
        session.execute(update(Item).where(Item.id == id).values(carry_num=carry_num))
    else:
        session.execute(
            update(Item).where(Item.id == id).values(carry_num=carry_num, is_carry=0)
        )
    session.commit()



from meterviewer import types as T
import cv2


# draw rectangle on image
def draw_rectangle(im: T.NpImage, rect: T.Rect) -> T.NpImage:
    """draw rectangle on image"""
    x0, y0, x1, y1 = rect["xmin"], rect["ymin"], rect["xmax"], rect["ymax"]
    x0, y0, x1, y1 = int(x0), int(y0), int(x1), int(y1)
    writable_im = im.copy()
    cv2.rectangle(writable_im, (x0, y0), (x1, y1), (0, 255, 0), 2)
    return writable_im


# write text on image
def draw_text(
    im: T.NpImage,
    text: str,
) -> T.NpImage:
    """write text on image"""
    x, y = 0, 30
    writable_im = im.copy()
    cv2.putText(writable_im, text, (x, y), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 255, 0), 2)
    return writable_im



import numpy as np

from . import cmp


def test_two_data():
    im1 = np.random.rand(100,100,3) * 255
    im2 = im1.copy()

    im3 = np.random.rand(100,100,3) * 255
    assert cmp.comp_ims(im1, im2)
    assert not cmp.comp_ims(im1, im3)

def test_one_chanel():
    im1 = np.random.rand(100,100,1) * 255
    im2 = im1.copy()

    # imarray = np.random.rand(100,100,1) * 255
    assert cmp.comp_ims(im1, im2)


from __future__ import annotations
import pathlib
import torch
from torch.utils.data import Dataset
import typing as t
import numpy as np
from torchvision import datasets
from PIL import Image
# from torchvision.transforms import transforms


class NumpyDataset(Dataset):
    def __init__(
        self,
        root: pathlib.Path,
        x_name: str | pathlib.Path,
        y_name: str | pathlib.Path,
        transform: t.Optional[t.Callable] = None,
        target_transform: t.Optional[t.Callable] = None,
    ):
        self.x_name = x_name
        self.y_name = y_name
        self.x = np.load(root / x_name)
        self.y = np.load(root / y_name)
        self.transform = transform
        self.target_transform = target_transform

    def __getitem__(self, index: int) -> tuple[Image.Image, np.ndarray]:
        x = self.x[index]
        y = self.y[index]

        x = Image.fromarray(x)

        if self.transform:
            x = self.transform(x)
        if self.target_transform:
            y = self.target_transform(y)
        return x, y



import pathlib
import typing as t
from meterviewer import T
from ..img import process
from matplotlib import pyplot as plt


def view_dataset_on_disk(name: str):
    def mmm(
        prefix_name: pathlib.Path,
        load_from_disk: t.Callable,
        view_dataset: t.Callable,
        show: bool = True,
        nums: int = 3,
    ):
        if not show:
            return

        assert prefix_name.exists()
        x = load_from_disk(prefix_name / name)
        x = process.np_to_img(x)
        view_dataset(nums, x)

    return mmm


def view_dataset_in_rows(num: int, imglist: T.ImgList):
    # 创建一个 GridSpec 实例，1行3列
    from matplotlib.gridspec import GridSpec

    fig = plt.figure(figsize=(15, 5))
    gs = GridSpec(1, 3, figure=fig)

    # 第一个子图
    ax1 = fig.add_subplot(gs[0, 0])
    ax1.imshow(imglist[0])

    # 第二个子图
    ax2 = fig.add_subplot(gs[0, 1])
    ax2.imshow(imglist[1])

    # 第三个子图
    ax3 = fig.add_subplot(gs[0, 2])
    ax3.imshow(imglist[2])

    plt.tight_layout()
    plt.show()


def view_dataset(num: int, imglist: T.ImgList):
    if imgshape := len(imglist[0].shape) > 3:
        raise Exception(f"imglist or Dataset meets error, shape: {imgshape}")
    for im in imglist[:num]:
        plt.imshow(im)
        plt.show()



"""根据数值判断进位情况；与数值相关的函数"""

import typing as t
from . import types as T


def is_carry(val: str) -> bool:
    return val[-1] in ("0", "9")


def is_carry_over(digit: T.DigitInt) -> bool:
    if digit[-1] == 0 or digit[-1] == 9:
        return True
    return False


def make_full(arr: t.List, length: int):
    new_arr = arr.copy()
    if len(arr) <= length:
        new_arr = [False] * (length - len(arr)) + new_arr
    else:
        raise ValueError(f"length is too short, len(arr) == {len(arr)}")
    return new_arr


def get_carry_array(digit: T.DigitStr) -> t.List[bool]:
    """get digit array from given digit string"""
    assert len(digit) >= 5
    res = []
    # 不是进位
    if not is_carry("".join(digit)):
        return make_full(res, len(digit))
    else:
        res.append(True)  # last digit is 0 or 9
        for c in reversed(digit[:-1]):
            res.append(True)
            if not c == "9":  # not continue when not 9
                break

    res = make_full(res, len(digit))
    return res


# def label_check(label: T.Label):
#     pass



# rotate image.
import os
import pathlib
import typing as t
import streamlit as st
import numpy as np
from meterviewer.config import get_root_path
from tmd.dataset import NumpyDataset

import torch
from torchvision.transforms import v2

torch.manual_seed(17)


def trans_function():
    trans = torch.nn.Sequential(
        v2.Pad(10, fill=0, padding_mode="constant"),
        v2.RandomRotation(30),
        # v2.RandomResizedCrop(size=(224, 224), antialias=True),
        v2.RandomHorizontalFlip(p=0.5),
        # transforms.CenterCrop(10),
        v2.ColorJitter(),
    )
    return trans


def rotate():
    st.title("rotate the image.")
    root_path = get_root_path()
    root_path = st.text_input("Enter the root path", value=str(root_path))
    root_path = pathlib.Path(root_path)

    path = st.text_input("Enter the path of the data", value="generated_merged")
    st.text(os.listdir(root_path / path))

    num = st.text_input("Enter the number of data")
    x_name = st.text_input("Enter the value of x_name", value="x_test.npy")
    y_name = st.text_input("Enter the value of y_name", value="y_test.npy")

    try:
        num_v: int = int(num)
    except ValueError:
        st.text("Please enter a number")

    folder_path = root_path / path
    x_path = folder_path / x_name
    y_path = folder_path / y_name

    status = st.text("loading dataset...")
    ds = NumpyDataset(
        folder_path,
        x_path,
        y_path,
        transform=trans_function(),
    )
    status.text("Done.")

    if st.button("Transform"):
        x, y = ds[num_v]
        st.text(f"x_shape: {x.size}, y_shape: {y.shape}")

        st.image(x, caption=f"Meterdata {num}")
        st.text(f"Meterdata {num} is {y}")


rotate()



from __future__ import annotations
import typing as t

import numpy as np
from PIL import Image

from meterviewer import types as T


def resize_img(img: T.NpImage, size: T.ImgSize) -> T.NpImage:
    # resize will reverse the size. height -> weight, weight -> height
    im = np.asarray(Image.fromarray(img).resize(list(reversed(size))), dtype=np.uint8)
    assert list(im.shape[:2]) == list(size), (im.shape, size)
    return im


def resize_imglist(
    imglist: T.ImgList,
    size: t.Optional[T.ImgSize] = None,
) -> T.ImgList:
    if not size:
        size = list(imglist[0].shape[:2])
    return [resize_img(img, size) for img in imglist]


def check_img_size(
    img: T.NpImage, size: t.Tuple[int, int], then: t.Callable[[T.NpImage], t.Any]
) -> bool | t.Any:
    if img.shape == size:
        return then(img)
    return False


def size_check(img_list: t.List[T.NpImage], size: t.Optional[t.List[int]] = None):
    assert img_list != [], "img_list should not be empty"
    if not size:
        size = list(img_list[0].shape)

    for i, img in enumerate(img_list):
        if list(img.shape) != size:
            raise ValueError(f"image: {i} size: {img.shape}, not match size: {size}")



import numpy as np
from matplotlib import pyplot as plt


def visualize_check(img, title):
    # 显示一个带确认按钮的窗口，如果确认则返回True，否则返回False
    pass


def show_img(img):
    # TODO: img valid test. use a gui to validate.
    is_show = 1
    if is_show:
        plt.imshow(img)
        plt.show()


def gen_img(size=(35, 25, 3)):
    return np.random.randint(0, 255, size=size, dtype=np.uint8)



import streamlit as st
import pandas as pd
import numpy as np
from matplotlib import pyplot as plt
from meterviewer.config import get_root_path
from meterviewer.datasets import imgv
from meterviewer.img import draw, cut
import pathlib


def example():
    st.title("Uber pickups in NYC")
    DATE_COLUMN = "date/time"
    DATA_URL = (
        "https://s3-us-west-2.amazonaws.com/"
        "streamlit-demo-data/uber-raw-data-sep14.csv.gz"
    )

    @st.cache_data
    def load_data(nrows):
        data = pd.read_csv(DATA_URL, nrows=nrows)
        lowercase = lambda x: str(x).lower()
        data.rename(lowercase, axis="columns", inplace=True)
        data[DATE_COLUMN] = pd.to_datetime(data[DATE_COLUMN])
        return data

    data_load_state = st.text("Loading data...")
    data = load_data(10000)
    data_load_state.text("Done! (using st.cache_data)")

    if st.checkbox("Show raw data"):
        st.subheader("Raw data")
        st.write(data)

    st.subheader("Number of pickups by hour")
    hist_values = np.histogram(data[DATE_COLUMN].dt.hour, bins=24, range=(0, 24))[0]
    st.bar_chart(hist_values)

    # Some number in the range 0-23
    hour_to_filter = st.slider("hour", 0, 23, 17)
    filtered_data = data[data[DATE_COLUMN].dt.hour == hour_to_filter]

    st.subheader("Map of all pickups at %s:00" % hour_to_filter)
    st.map(filtered_data)


import typing as t

global_func: t.Dict[str, t.Any] = {}


def use_func(name: str):
    f = global_func.get(name, None)
    if not f:
        raise ValueError(f"func {name} not found")


def must_str(object) -> str:
    return str(object)


def must_loop(arr: t.Iterable, func: t.Callable, custom_error):
    """raise custom error if not looped"""
    is_loop, set_loop = looped()

    for element in arr:
        func(element)
        set_loop()

    if not is_loop():
        raise custom_error("must loop")


def looped():
    loop = False

    def is_loop():
        return loop

    def set_loop():
        nonlocal loop
        loop = True

    return is_loop, set_loop


Result = t.TypeVar("Result")


def try_again(
    total: int,
    retry_f: t.Callable[[], Result],
    is_validate_func: t.Callable[[t.Any], bool],
    fail_message: str,
) -> Result:
    i = 0
    res = retry_f()
    while not is_validate_func(res):
        res = retry_f()
        i += 1
        if i == total:
            raise Exception(f"Cannot find valid result, {fail_message}")

    return res



import streamlit as st
import pandas as pd
import pathlib
import numpy as np
import os
import ast
import shutil


def main():
    # st.Page("copy page", title="Tool to backup numpy")
    st.title("Tool to backup numpy.")
    root_path = pathlib.Path("/home/xiuhao/work/Dataset/HRCdata")
    root_path = st.text_input("Enter the root path", value=str(root_path))
    root_path = pathlib.Path(root_path)

    folder = st.text_input("Enter the path of the data", value="Train")

    st.text(os.listdir(root_path / folder))

    x_name = st.text_input("Enter the value of x_name", value="x_train.npy")
    y_name = st.text_input("Enter the value of y_name", value="y_train.npy")

    new_x_name = st.text_input(
        "Enter the value of new_x_name", value="x_train.original.npy"
    )
    new_y_name = st.text_input(
        "Enter the value of new_x_name", value="y_train.original.npy"
    )

    if st.button("Confirm"):
        shutil.copy(root_path / folder / x_name, root_path / folder / new_x_name)
        shutil.copy(root_path / folder / y_name, root_path / folder / new_y_name)


main()



from PIL import Image
import streamlit as st
import pandas as pd
import pathlib
import numpy as np
from meterviewer.datasets import imgv
from matplotlib import pyplot as plt


def cmd():
    st.Page("compare page", title="Tool to compare images")
    root_path = pathlib.Path("/home/xiuhao/work/Dataset/HRCdata")
    dataset_1, dataset_2 = "Test", "Train"
    num = 1
    x1_name, x2_name = "x_all.npy", "x_train.npy"
    x1_path = root_path / dataset_1 / x1_name
    x2_path = root_path / dataset_2 / x2_name
    x1 = np.load(x1_path)
    x2: np.ndarray = np.load(x2_path)

    def save_img(path, arr):
        im = Image.fromarray(arr)
        im.save(path)

    im = x1[num]
    save_img("/tmp/im1.png", im)
    print("example saved.")

    inx = imgv.find_images(x2)
    if inx == -1:
        print("No same image found.")
    else:
        print(inx)
        print("Same image found.")
        save_img("/tmp/im2.png", x2[inx])


def main():
    st.title("Tool to compare images.")
    root_path = pathlib.Path("/home/xiuhao/work/Dataset/HRCdata")
    root_path = st.text_input("Enter the root path", value=str(root_path))
    root_path = pathlib.Path(root_path)

    dataset_1 = st.text_input("Enter the dataset 1 name", value="Test")
    dataset_2 = st.text_input("Enter the dataset 2 name", value="Train")

    num = st.text_input("Enter the number of data")
    x1_name = st.text_input("Enter the value of x1_name", value="x_all.npy")
    # y_name = st.text_input("Enter the value of y_name", value="y_test.npy")
    x2_name = st.text_input("Enter the value of x2_name", value="x_train.npy")

    try:
        data_load_state = st.text("reading data...")
        num = int(num)
        data_load_state.text("Done!")
    except ValueError:
        data_load_state.text("Please enter a number")
        return

    # folder_path = root_path / path
    x1_path = root_path / dataset_1 / x1_name
    x2_path = root_path / dataset_2 / x2_name

    assert x1_path.exists()
    assert x2_path.exists()

    x1 = np.load(x1_path)
    x2: np.ndarray = np.load(x2_path)

    st.text(f"x1_shape: {x1.shape}, x2_shape: {x2.shape}")

    st.text("Selected image: ")
    st.image(x1[num], caption=f"Meterdata {num}")

    # raise Exception("Not implemented yet")

    im = x1[num]
    if st.button("Search"):
        state = st.text("searching")
        inx = imgv.find_images(im, x2.tolist())
        if inx == -1:
            st.text("No same image found.")
        else:
            st.text("Same image found.")
            st.image(x2[inx], caption=f"Meterdata {inx}")
        state.text("finished")


if __name__ == "__main__":
    # cmd()
    main()



import pathlib
from meterviewer.models import func
import typing as t


def from_db() -> t.Generator[t.Tuple[pathlib.Path, pathlib.Path, str], None, None]:
    dbpath = "alldata.db"
    items = func.get_carry_items(dbpath)
    items = list(items)
    for item in items:
        yield (
            pathlib.Path(item.filename),
            pathlib.Path(item.filename).parent,
            str(id),
        )



# generate single digit
from __future__ import annotations

import pathlib
import typing as t
from datetime import datetime
from pathlib import Path as P

import numpy as np

from meterviewer import T, files
from meterviewer.datasets import dataset
from meterviewer.img import process


def use_write(
    im: T.NpImage, filename: str
) -> t.Callable[[P, int, t.Literal["normal", "digit"]], None]:
    """
    Create a function that writes an image to a specific location.

    Args:
        im (T.NpImage): The image to be written.
        filename (str): The name of the file to be created.

    Returns:
        Callable: A function that writes the image to a specified path.
    """

    def write_to(
        root_path: P,
        digit: int,
        type_: t.Literal["normal", "digit"],
    ):
        """
        Write the image to a specific location based on the provided parameters.

        Args:
            root_path (P): The root path where the dataset will be created.
            digit (int): The digit represented in the image.
            type_ (Literal["normal", "digit"]): The type of image (normal or digit).
        """
        # Get current date for dataset naming
        date = datetime.now().strftime("%Y-%m-%d")
        dataset_path = root_path / f"generated-{date}"
        digit_path = dataset_path / type_ / str(digit)

        # Create directory if it doesn't exist
        digit_path.mkdir(parents=True, exist_ok=True)

        # Save the image as a numpy array
        np.save(digit_path / filename, im)

    return write_to


def generate_total_dataset(
    length: int,
    size: int,
    create_np: t.Callable[[T.ImgList, t.List[t.List[int]]], None],
):
    """size: dataset samples size"""

    def get_random_number():
        pass

    def get_one_img(digit: int):
        pass

    def join_with(imglist: T.ImgList):
        # this join-with will check the size
        return process.join_img(imglist, process.size_check)

    def multiple_gen(get_random_number: t.Callable, get_one_img: t.Callable):
        img_list = []
        v_list = []
        for _ in range(size):
            gen_func = generate_block(length, get_random_number)
            im, v = gen_func(get_one_img=get_one_img, join_img=join_with)
            img_list.append(im)
            v_list.append(v)

        return create_np(img_list, v_list)

    return multiple_gen


def generate_block(length: int, get_random_number: t.Callable[[int], t.List[int]]):
    def gen(
        get_one_img: t.Callable[[int], T.NpImage],
        join_img: t.Callable[[T.ImgList], T.NpImage],
    ):
        """generate single digit image"""

        number: t.List[int] = get_random_number(length)
        assert len(number) == length

        img_list = []
        for digit in number:
            img = get_one_img(digit)
            img_list.append(img)

        im = join_img(img_list)
        return im, number

    return gen


def img_selector(root_path: P):
    def get_random_img(
        dataset_name: str,
    ) -> t.Tuple[T.ImgList, t.List[str], t.Literal["normal", "digit"]]:
        dp = dataset.get_dataset_path(root_path, dataset_name)

    return get_random_img


def generate_single(
    root_path: pathlib.Path,
    get_random_img: t.Callable[
        [str], t.Tuple[T.ImgList, t.List[str], t.Literal["normal", "digit"]]
    ],
):
    """generate single digit image, split to normal and digit folder"""

    def handle_img(
        im: T.NpImage,
        filename: str,
        digit: int,
        type_: t.Literal["normal", "digit"],
    ):
        """write to specific folder"""
        write = use_write(im, filename)
        write(root_path, digit, type_)

    def inner(filename):
        def handle_dataset(dataset_name: str | pathlib.Path):
            imgs, values, types_ = get_random_img(str(dataset_name))
            for im, v in zip(imgs, values):
                handle_img(im, v, filename, types_)

        dataset.handle_datasets(root_path, handle_dataset)

    return inner



import pathlib
import typing as t
import numpy as np
from meterviewer import files
from .utils import gen_img


def test_md5():
    res = files.compute_md5("src/tests/for_md5")
    assert res == "5093b7d2c218eaebde04b58b20491ef6"


def test_reshape():
    img = gen_img(size=(35, 25, 3))
    assert files.transform_img(img).shape == (1, 35, 25, 3)

    # test concat
    im1 = files.transform_img(img)
    im2 = files.transform_img(img)
    assert np.vstack([im1, im2, im2, im2]).shape == (4, 35, 25, 3)


def test_transform_label():
    label = ["1", "2", "3", "4", "5"]
    assert files.transform_label(label).shape == (1, 5)


def test_scan_pics(root_path):
    path = root_path / pathlib.Path(r"lens_6/XL/XL/M1L1XL/Digit/0")
    count = 0
    pics: t.List[pathlib.Path] = []

    def update_count(name):
        nonlocal count, pics
        pics.append(name)
        count += 1

    [update_count(name) for name in files.scan_pics(path)]
    assert count == 20
    assert pics[0].parent.name == "0"
    # path like this:
    # PosixPath('/home/svtter/Work/Dataset/MeterData/lens_6/XL/XL/M1L1XL/Digit/0/4_2018-11-23-19-21-01.jpg')
    # assert False, pics



# dataset name.
import typing as t
from meterviewer import types as T


def normal_x_y() -> t.Tuple[str, str]:
    return "x.npy", "y.npy"


def train_x_y() -> t.Tuple[str, str]:
    return T.x_name, T.y_name


name_funcs = {
    "normal": normal_x_y,
    "train": train_x_y,
}



import typing as t
import functools
from meterviewer import T
from meterviewer.img import process
from loguru import logger


def join_with_fix(
    imglist: T.ImgList, check_func: t.Callable, fix_func: t.Callable
) -> T.NpImage:
    """修饰 join_func"""
    # merge images horizontally
    try:
        return process.join_img(imglist, check_func)
    except ValueError as e:
        logger.debug(e)
        imglist = fix_func(imglist)
    return process.join_img(imglist, check_func)


join_with_resize: T.JoinFunc = functools.partial(
    join_with_fix, fix_func=process.resize_imglist
)



"""view image and value"""

import pathlib
import typing as t
import numpy as np
from pathlib import Path as P

from tqdm import tqdm
from loguru import logger
from matplotlib import pyplot as plt
from PIL import ImageFile

from meterviewer import T
from meterviewer.datasets import config
from meterviewer.img import cmp

logger.add("./logs/meterviewer-proc.log")

# 修复 PIL 无法读取截断图片的问题; matplotlib 底层居然是 PIL？是我搞错了吗？
ImageFile.LOAD_TRUNCATED_IMAGES = True


def view_one_img_v(filepath: P) -> t.Tuple[T.NpImage, str, T.Rect]:
    """同时浏览图片和值"""
    im = plt.imread(str(filepath))
    v, rect = config.get_xml_config(filepath)
    return (im, v, rect)


hash_store = {}


def save_hash(im: T.NpImage, index: int):
    hash = cmp.get_hash(im)
    res = hash_store.get(hash, None)
    if res is None:
        hash_store[hash] = index
    else:
        logger.error(f"Hash: {hash} is same with {res} and {index}")
        raise Exception("Img should not be same.")


def find_images(img_list: T.ImgList) -> int:
    logger.debug("start finding...")
    for i, img in tqdm(enumerate(img_list)):
        try:
            save_hash(img, i)
        except Exception as e:
            logger.warning(f"warning: {e} in {i}")
            return i
    # index in the np_path
    return -1


def find_hash_in_numpy(im: T.NpImage, np_path: pathlib.Path) -> int:
    save_hash(im, -1)
    x = np.load(np_path)
    res = find_images(list(x))
    return res



from string import Template


def dataset():
    data = {}
    str_template = Template("$name_$peroid.npy")

    names = ["x", "y"]
    peroids = ["train", "test"]

    s = str_template.safe_substitute(name="train", peroid="2021")

    def get_root_path():
        pass



import pathlib
import typing as t
import numpy as np
from . import types as T
import hashlib

import toml


def use_smart_name(dataset: pathlib.Path) -> T.NameFunc:
    """
    Determine the appropriate file names for X and Y data in a dataset.
    Returns a function that provides the correct file names.
    """
    x_name_list = ["x_train.npy", "x_all.npy", "x.npy"]
    y_name_list = ["y_train.npy", "y_all.npy", "y.npy"]

    for x, y in zip(x_name_list, y_name_list):
        if (dataset / pathlib.Path(x)).exists() and (
            dataset / pathlib.Path(y)
        ).exists():

            def get_x_y():
                return x, y

            return get_x_y
    else:
        raise Exception("No valid npy files found")


def compute_md5(file_path, chunk_size=8192):
    """
    Compute the MD5 hash of a file.

    Args:
        file_path (str): The path to the file.
        chunk_size (int): The size of each chunk to read from the file.

    Returns:
        str: The MD5 hash of the file in hexadecimal format.
    """
    md5_hash = hashlib.md5()

    with open(file_path, "rb") as f:
        while chunk := f.read(chunk_size):
            md5_hash.update(chunk)

    return md5_hash.hexdigest()


def read_toml(filename: pathlib.Path) -> t.Optional[t.Dict]:
    """
    Read and parse a TOML file, returning its contents as a dictionary.
    """
    try:
        with open(filename, "r") as f:
            data = toml.load(f)
            print(f"Read '{filename}' successfully.")
            return data
    except Exception as e:
        print(f"Error to read '{filename}': {e}")


def write_toml(filename: pathlib.Path, data):
    """
    Write data to a TOML file.
    """
    try:
        with open(filename, "w") as f:
            toml.dump(data, f)
        print(f"Data written to '{filename}' successfully.")
    except Exception as e:
        print(f"Error writing to '{filename}': {e}")


def write_shape(img: T.ImgList, nums: int = 3):
    """
    Write the shapes of the first 'nums' images to a debug log file.
    """
    debug_file = pathlib.Path("debug.log")
    with open(debug_file, "w") as f:
        for i in range(nums):
            f.write(f"img_{i} shape is: {img[i].shape}\n")


def scan_pics(path: pathlib.Path) -> t.Iterator[pathlib.Path]:
    """
    Scan a directory for image files (jpg, png, jpeg) and yield their paths.
    """
    for p in path.iterdir():
        if p.suffix in [".jpg", ".png", ".jpeg"]:
            yield p


def transform_img(img: T.NpImage) -> T.NpImage:
    """
    Reshape a single image by adding a batch dimension.
    """
    return np.expand_dims(img, axis=0)


def transform_label(label: T.DigitStr, to_int: bool) -> T.Label:
    """
    Transform a label string into a numpy array, optionally converting to integers.
    """
    if to_int:
        label_ = [int(i) for i in label]
    else:
        label_ = label
    return np.expand_dims(np.array(label_), axis=0)


def save_imgs_labels(
    imgs_labels: t.Tuple[t.List[T.NpImage], t.List[T.DigitStr]],
    np_name: t.Callable[[], t.Tuple[pathlib.Path, str, str]],
    save_to_disk: t.Callable,
):
    """
    Save a list of images and labels to disk as numpy arrays.
    """
    imgs, labels = imgs_labels
    imgs = [np.expand_dims(img, axis=0) for img in imgs]
    labels_ = [transform_label(label, False) for label in labels]
    x_train = np.vstack(imgs)
    y_train = np.vstack(labels_)

    prefix_path, x_name, y_name = np_name()
    save_to_disk(str(prefix_path / x_name), x_train)
    save_to_disk(str(prefix_path / y_name), y_train)


def save_img_labels_with_default(
    imgs: t.List[T.NpImage],
    labels: t.List[T.DigitStr],
    prefix_name: pathlib.Path,
    save_to_disk: t.Callable,
):
    """
    Save images and labels to disk using default file names.
    """
    def np_name():
        return prefix_name, T.x_name, T.y_name

    return save_imgs_labels(
        (imgs, labels),
        np_name,
        save_to_disk,
    )


def save_to_disk(filename: str, data: np.ndarray):
    """
    Save a numpy array to disk.
    """
    with open(filename, "wb") as f:
        np.save(f, data)


def load_from_disk(filename) -> np.ndarray:
    """
    Load a numpy array from disk.
    """
    with open(filename, "rb") as f:
        return np.load(f)


def load_from_disk_with_md5(filename, with_md5: str) -> t.Tuple[np.ndarray, str]:
    """
    Load a numpy array from disk and optionally compute its MD5 hash.
    """
    md5 = ""
    if with_md5:
        md5 = compute_md5(filename)
    return load_from_disk(filename), md5



from __future__ import annotations
import typing as t
import pathlib
from meterviewer import files, types as T
from meterviewer.datasets import dataset


def view_merge_np(
    current_dataset: str | pathlib.Path,
    view_dataset: t.Callable[[int, T.ImgList], None] = dataset.view_dataset,
    get_x_y: T.NameFunc = lambda: (T.x_name, T.y_name),
):
    """view already handled data."""
    pp = pathlib.Path(current_dataset)

    x_name, _ = get_x_y()
    view_func = dataset.view_dataset_on_disk(x_name)

    view_func(
        prefix_name=pp,
        view_dataset=view_dataset,
        load_from_disk=files.load_from_disk,
    )


def read_details(current_dataset: str) -> t.Optional[t.Dict]:
    return files.read_toml(pathlib.Path(current_dataset) / "details.gen.toml")


def get_x_y_name() -> t.Tuple[str, str]:
    return T.x_name, T.y_name


def write_details(
    current_dataset: str | pathlib.Path,
    get_xy_name: T.NameFunc = get_x_y_name,
):
    pp = pathlib.Path(current_dataset)

    def write_to_file(details, overwrite=True):
        p = pp / "details.gen.toml"
        if not overwrite and p.exists():
            print("Failed to write, file exists")
            return
        return files.write_toml(p, details)

    x_name, y_name = get_xy_name()

    dataset.show_details(
        get_x_train=lambda: files.load_from_disk(pp / x_name),
        get_y_train=lambda: files.load_from_disk(pp / y_name),
        get_details=lambda x, y: dataset.get_details(pp, x, y),
        write_to_file=write_to_file,
    )



"""generate database"""

import typing as t
from pathlib import Path as P

from meterviewer import T, files
from meterviewer.datasets import dataset, imgv, single
from meterviewer.models import littledb
from meterviewer.values import is_carry

# Type hint for database insertion function
dbInsertFunc = t.Callable[[str, int, bool], None]


def generate_for_one_dataset(dataset: P, insert: dbInsertFunc):
    """
    Process a single dataset and insert its image data into the database.

    Args:
        dataset (Path): Absolute path to the dataset directory
        insert (dbInsertFunc): Callback function to insert data into the database
            Expected signature: (image_path: str, value: int, is_carry: bool) -> None

    Raises:
        AssertionError: If dataset path is not absolute
    """
    assert dataset.is_absolute()
    for img_file in files.scan_pics(dataset):
        _, v, _ = imgv.view_one_img_v(img_file)
        # transform str to int
        val = int(v)

        insert(str(img_file), val, is_carry(v))


def generate_db_for_all(root: P, db_path: P):
    """
    Generate a single database containing data from all datasets.

    Args:
        root (Path): Root directory containing all datasets
        db_path (Path): Relative path where the database should be created

    Raises:
        AssertionError: If db_path is absolute
    """
    # insert to one database.
    assert not db_path.is_absolute()

    insert, _ = littledb.create_db(str(db_path))

    def generate_one(dataset: P):
        return generate_for_one_dataset(
            single.get_dataset_path(root, str(dataset)), insert
        )

    dataset.handle_datasets(
        root,
        generate_one,
    )


def create_db(root_path: P):
    """
    Create separate databases for each dataset in the root directory.
    Each database will be named 'items.db' and placed in its respective dataset directory.

    Args:
        root_path (Path): Root directory containing all datasets
    """
    db_name = "items.db"

    def handle_dataset(dataset_name: P):
        dataset_path = dataset.get_dataset_path(root_path, str(dataset_name))
        littledb.create_db(str(dataset_path / db_name))

    dataset.handle_datasets(
        root_path,
        handle_func=handle_dataset,
    )



# read dataset from file system.
from pathlib import Path as P
from dataclasses import dataclass
import xml.etree.ElementTree as ET
import typing as t

from meterviewer import types as T, func as F


@dataclass
class RectO(object):
    xmin: str = ""
    ymin: str = ""
    xmax: str = ""
    ymax: str = ""

    def check(self):
        # assert rect.xmin != "" and rect.ymin != "" and rect.xmax != "" and rect.ymax != "", rect
        assert self.xmin != "" and self.ymin != "" and self.xmax != "" and self.ymax != "", self

    def to_dict(self) -> T.Rect:
        return {
            "xmin": self.xmin,
            "ymin": self.ymin,
            "xmax": self.xmax,
            "ymax": self.ymax,
        }

    def __str__(self):
        return f"RectO({self.xmin}, {self.ymin}, {self.xmax}, {self.ymax})"

    def __repr__(self):
        return self.__str__()


def read_xml(filename: P, read_func: t.Callable[[t.Any], t.Any]):
    # 解析 XML 文件
    tree = ET.parse(filename)
    root = tree.getroot()
    return read_func(root)


def read_rect_from_node(root: t.Iterable) -> t.Tuple[str, T.Rect]:
    val, rect_dict = "", RectO()
    for child in root:
        # find object
        if not child.tag == "object":
            continue
        for subchild in child:
            if subchild.tag == "name":
                val = F.must_str(subchild.text)

            # print(subchild.tag, subchild.text)
            if subchild.tag == "bndbox":
                for sub in subchild:
                    setattr(rect_dict, sub.tag, sub.text)

    return val, rect_dict.to_dict()


def read_single_digit_rect(filename) -> t.List[RectO]:
    def func(root: t.Iterable) -> t.List[RectO]:
        def find_no(node) -> t.Tuple[int, RectO]:
            no = -1
            rect = RectO()
            for subchild in node:
                if subchild.tag == "no":
                    no = int(subchild.text.strip())
                else:
                    rect = set_rect(rect, subchild)
            rect.check()
            assert no != -1, "cannot find no number"
            return no, rect

        def set_rect(rect: RectO, sub) -> RectO:
            assert hasattr(rect, sub.tag), (sub.tag, sub.text)
            setattr(rect, sub.tag, sub.text)
            assert getattr(rect, sub.tag) != ""
            return rect

        def num_check():
            seta = {0, 1, 2, 3, 4, 5}
            setb = set()

            def is_valid():
                return seta == setb, (seta, setb)

            def set_num(no):
                setb.add(no)

            return set_num, is_valid

        digit_rect = [RectO() for _ in range(6)]
        set_num, is_valid = num_check()
        is_loop, set_loop = F.looped()

        for child in root:
            if child.tag == "digit":
                no, rect = find_no(child)
                _ = set_loop(), set_num(no)
                digit_rect[no] = rect

        assert is_loop(), "node has no child"
        cond, _ = is_valid()
        assert cond

        # return digit_rect, root, find_no, find_rect
        return digit_rect

    return read_xml(filename, func)


def get_single_digit_values(filename: P) -> t.Tuple[str, T.Rect]:
    val, _ = read_xml(filename, read_rect_from_node)
    block_pos = read_xml(filename, read_single_digit_rect)
    return val, block_pos


typeOfrect = t.Literal["single", "block"]


def read_rect_from_file(xml_path: P, type_: typeOfrect):
    assert xml_path.suffix == ".xml"
    function_map: t.Mapping[typeOfrect, t.Callable[[P], t.Any]] = {
        "single": read_single_digit_rect,
        "block": get_rectangle,
    }
    func = function_map[type_]
    return func(xml_path)


def get_rectangle(filename: P) -> T.Rect:
    _, rect = read_xml(filename, read_rect_from_node)
    return rect


def get_xml_config_path(img_path: P, types: t.Literal["value", "block", "single"] = "value") -> P:
    """filename (test.png or test.jpg) to config (res.xml or suffix.xml)"""
    dataset_path = img_path.parent

    def value_path():
        config_p = P(dataset_path) / "baocun"
        assert img_path.suffix in (".jpg", ".jpeg")
        filename = img_path.stem + ".xml"
        return config_p / filename

    def block_path():
        return P(dataset_path) / "config" / "block.xml"

    def single_path():
        return P(dataset_path) / "config" / "res.xml"

    funcs: t.Mapping[str, t.Callable] = {
        "value": value_path,
        "block": block_path,
        "single": single_path,
    }

    return funcs[types]()


def get_xml_config(img_path: P) -> t.Tuple[str, T.Rect]:
    return read_xml(get_xml_config_path(img_path), read_rect_from_node)



# process image, to fit the training proposals.

from __future__ import annotations

import sys
import typing as t

import numpy as np
from matplotlib import pyplot as plt

# from matplotlib import pyplot as plt
from .. import types as T
from .resize import check_img_size, resize_img, resize_imglist, size_check  # noqa


def np_to_img(data: np.ndarray) -> T.ImgList:
    return list(data)


def join_img(
    imglist: T.ImgList,
    check_func: t.Callable[[t.Any], t.Any],
) -> T.NpImage:
    # merge images vertically
    check_func(imglist)
    return np.hstack(imglist)


def get_random_img(num: int, img_from: t.Callable) -> T.NpImage:
    """get random img
    num: digit num of img
    """
    get_img = img_from()
    return get_img(num)


def img_from(folder: str = ""):
    # open folder to get all images.
    def get_img(num):
        return np.random.randint(1, 255, size=(10, 20))

    return get_img


def get_img_list(nums: t.List[int]) -> t.List[T.NpImage]:
    imgs = []
    for i in nums:
        imgs.append(get_random_img(int(i), lambda: None))
    return imgs


def number_to_string(number: int, length: int) -> t.List[str]:
    # create a string list from number, with fixed length.
    return list(str(number).zfill(length))


def empty_check(*args, **kwargs):
    """do nothing function"""
    pass


def gen_block_img(number: int, length: int):
    """inside memory to generate, use as a sample function."""
    num_l = [int(i) for i in number_to_string(number, length)]
    return join_img(get_img_list(num_l), empty_check)


def show_img(img, is_stop):
    plt.imshow(img)
    plt.show()
    if is_stop:
        sys.exit(-1)


def gen_empty_im(size: t.Tuple[int, int, int]):
    return np.zeros(size, dtype=np.uint8)



# create a little db for storing the images and values

from __future__ import annotations
import typing as t
import pathlib
from typing import List
from typing import Optional
from sqlalchemy import ForeignKey
from sqlalchemy import String
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.orm import Mapped, Session
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import relationship
from sqlalchemy import create_engine


def get_session(filepath: str | pathlib.Path):
    filepath = str(filepath)
    engine = create_engine(f"sqlite:///{filepath}", echo=True)
    with Session(engine) as session:
        return session


def create_db(filepath: str):
    engine = create_engine(f"sqlite:///{filepath}", echo=True)
    Base.metadata.create_all(engine)

    with Session(engine) as session:

        def insert_one(filename: str, value: int, is_carry: bool):
            item = Item(filename=filename, value=value, is_carry=is_carry)
            session.add(item)
            session.commit()

        def insert_all(items: t.List[Item]):
            session.add_all(items)
            session.commit()

        return insert_one, insert_all


class Base(DeclarativeBase):
    pass


class Item(Base):
    __tablename__ = "item"
    id: Mapped[int] = mapped_column(primary_key=True)
    filename: Mapped[str] = mapped_column(String(30))
    value: Mapped[int] = mapped_column()
    is_carry: Mapped[bool] = mapped_column()
    carry_num: Mapped[int] = mapped_column()

    def __repr__(self) -> str:
        return f"Item(id={self.id!r}, filename={self.filename!r}, value={self.value!r})"


# examples
def examples():
    class User(Base):
        __tablename__ = "user_account"
        id: Mapped[int] = mapped_column(primary_key=True)
        name: Mapped[str] = mapped_column(String(30))
        fullname: Mapped[Optional[str]]
        addresses: Mapped[List["Address"]] = relationship(
            back_populates="user", cascade="all, delete-orphan"
        )

        def __repr__(self) -> str:
            return (
                f"User(id={self.id!r}, name={self.name!r}, fullname={self.fullname!r})"
            )

    class Address(Base):
        __tablename__ = "address"
        id: Mapped[int] = mapped_column(primary_key=True)
        email_address: Mapped[str]
        user_id: Mapped[int] = mapped_column(ForeignKey("user_account.id"))
        user: Mapped["User"] = relationship(back_populates="addresses")

        def __repr__(self) -> str:
            return f"Address(id={self.id!r}, email_address={self.email_address!r})"



import abc


class Generator(abc.ABC):
    """所有生成器的入口，需要实现的生成函数"""

    @abc.abstractmethod
    def gen_func(self):
        pass


class Source(object):
    """"""

    pass


class ImageSource(Source):
    """原图像，定义清楚之后可以用于生成新的图像"""

    pass



from __future__ import annotations
import pytest
import functools
from meterviewer import files, types as T
import pathlib

from meterviewer.datasets import dataset, single
from meterviewer.img import process
from tests.utils import show_img


def test_create_label():
    assert dataset.create_str_with_fill(1234, 5, 8) == [
        "0",
        "1",
        "2",
        "3",
        "4",
        "x",
        "x",
        "x",
    ]
    assert dataset.create_str_with_fill(12340, 5, 8) == [
        "1",
        "2",
        "3",
        "4",
        "0",
        "x",
        "x",
        "x",
    ]


def test_python_syntax():
    arr = ["1", "2", "3", "4"]
    arr.extend(["x"] * 3)
    assert arr == list("1234xxx")
    assert arr == ["1", "2", "3", "4", "x", "x", "x"]

    arr = list("1234")
    arr.extend(["x"] * 3)
    assert arr == list("1234xxx")


def test_fill_digit():
    assert dataset.fill_digit(list("12312"), 7) == list("12312xx")
    assert dataset.fill_digit(list("01234"), 8) == list("01234xxx")
    assert dataset.fill_digit(["0", "1", "2", "3", "4"], 8) == [
        "0",
        "1",
        "2",
        "3",
        "4",
        "x",
        "x",
        "x",
    ]
    assert list("12312xx") == ["1", "2", "3", "1", "2", "x", "x"]
    assert list("1212xx") == ["1", "2", "1", "2", "x", "x"]
    assert list("0123") == ["0", "1", "2", "3"]

    with pytest.raises(ValueError):
        dataset.fill_digit(list("12312"), 4)
        dataset.fill_digit(list("12"), 4)


def test_view_on_disk(root_path):
    f = dataset.view_dataset_on_disk(T.x_name)
    f(
        prefix_name=root_path / r"generated",
        load_from_disk=files.load_from_disk,
        view_dataset=dataset.view_dataset,
        show=True,
    )


def test_create_dataset(root_path):
    P = functools.partial

    def read_rand_img(digit: int | str):
        return single.read_rand_img(
            digit=digit,
            root=root_path,
            get_dataset=lambda: "M1L1XL",
        )

    gen_block = P(
        dataset.generate_block_img,
        join_func=dataset.join_with_resize,
        read_rand_img=read_rand_img,
    )

    path = root_path / pathlib.Path(r"generated")
    path.mkdir(exist_ok=True)

    filesave = P(
        files.save_img_labels_with_default,
        prefix_name=path,
        save_to_disk=files.save_to_disk,
    )

    def check_imgs(imglist):
        size = imglist[0].shape
        show_img(imglist[0])
        imgs = process.resize_imglist(imglist)
        for im in imgs:
            show_img(im)
            assert size == im.shape

    imgs, labels = dataset.create_dataset_func(check_imgs=lambda x: None, total=9)(
        length=5,
        nums=10,
        gen_block_img=gen_block,
    )
    filesave(imgs=imgs, labels=labels)


def test_generate_block_img(root_path):
    def read_rand_img(digit: int | str):
        return single.read_rand_img(
            digit=digit,
            root=root_path,
            get_dataset=lambda: "M1L1XL",
        )

    im = dataset.generate_block_img(
        ["1", "2", "3", "4"],
        process.join_img,
        read_rand_img,
    )
    show_img(im)

    im = dataset.generate_block_img(
        ["1", "2", "3", "5", "6"],
        dataset.join_with_resize,
        read_rand_img,
    )
    show_img(im)

    im = dataset.generate_block_img(
        ["1", "2", "3", "5", "6", "x", "x"],
        dataset.join_with_resize,
        read_rand_img,
    )
    show_img(im)


def test_get_random_dataset(root_path):
    _, index = dataset.get_random_dataset(root_path, dataset.get_dataset_list)
    assert index in range(0, 74)


def test_read_random_img(root_path):
    im = single.read_rand_img(root_path, lambda: "M1L1XL", 5)
    show_img(im)

    im = single.read_rand_img(root_path, lambda: "M1L1XL", "x")
    show_img(im)


def test_read_random_digit(root_path):
    path_gen = single.read_single_digit(root_path, lambda: "M1L1XL", 0, promise=False)()
    p = next(path_gen)
    assert pathlib.Path(p).exists()


def test_dataset_list(root_path):
    # path = root_path / pathlib.Path("lens_6/XL/XL")
    count = 0
    pics = []

    def update_count(name):
        nonlocal count, pics
        pics.append(name)
        count += 1

    dataset.handle_datasets(root_path, update_count)
    assert count == 74
    # assert False, pics



import typing as t
import numpy as np


class Validator(object):
    # a image validator

    def is_img(self, img: np.ndarray) -> bool:
        conds = [
            lambda x: isinstance(x, np.ndarray),
            lambda x: len(x.shape) == 3,  # gray and color.
            lambda x: x.dtype == np.uint8,
            lambda x: x.shape[2] in [1, 3],  # gray and color.
            lambda x: x.shape[0] > 0 and x.shape[1] > 0,
        ]
        for c in conds:
            if not c(img):
                return False
        return True

    def valid_shape(self, img: np.ndarray, shape: t.Tuple) -> bool:
        return img.shape == shape



from meterviewer.img import process

# import functools
# from matplotlib import pyplot as plt
from tests.utils import show_img, gen_img


def test_resize_imglist():
    imglist = [gen_img(size=(35, 25, 3)), gen_img(size=(34, 25, 3))]
    process.resize_imglist(imglist, size=[35, 25])

    imglist = [gen_img(size=(31, 170, 3)), gen_img(size=(31, 175, 3))]
    process.resize_imglist(imglist)


def test_resize_img():
    im = gen_img(size=(35, 25, 3))
    process.resize_img(im, size=[35, 25])


def test_number_to_string():
    assert process.number_to_string(1, 5) == list("00001")
    assert process.number_to_string(10, 5) == ["0", "0", "0", "1", "0"]
    assert process.number_to_string(100, 5) == list("00100")
    assert process.number_to_string(1000, 5) == list("01000")


def test_join():
    imglist = [
        process.get_random_img(1, process.img_from),
        process.get_random_img(2, process.img_from),
        process.get_random_img(3, process.img_from),
    ]
    res = process.join_img(imglist, process.empty_check)
    assert res.shape[1] > res.shape[0]
    # seems correct
    show_img(res)



"""handle function based on single, that is dataset_name/[0-9] format"""

from __future__ import annotations

import functools
import pathlib
import random
import typing as t

from loguru import logger
from matplotlib import pyplot as plt

from meterviewer import T, files, func
from meterviewer.img import process

from .dataset import get_dataset_path


def path_fusion(
    root: pathlib.Path,
    dataset_name: str,
    num: int,
):
    """return single digit"""
    p = get_dataset_path(root, dataset_name) / "Digit" / str(num)
    return p


def read_rand_img(
    root: pathlib.Path,
    get_dataset: t.Callable[[], t.Union[str, pathlib.Path]],
    digit: t.Union[int, str],
    promise=False,
) -> T.NpImage:
    """return a random image, single digit."""
    if digit == "x":
        im = process.gen_empty_im((32, 40, 3))
        return im

    get_one = read_single_digit(
        root,
        get_dataset=get_dataset,
        num=int(digit),
        promise=promise,
    )
    all_imgs = list(get_one())
    length = len(all_imgs)

    if length == 0:
        raise Exception(f"Dataset contains no images, dataset: {get_dataset()}")

    i = random.randint(0, length - 1)
    im = plt.imread(all_imgs[i])
    return im


def read_single_digit(
    root_path: pathlib.Path,
    get_dataset: t.Callable[[], str | pathlib.Path],
    num: int,
    promise: bool,
) -> t.Callable[[], t.Iterator[pathlib.Path]]:
    """promised return"""
    assert num in range(0, 10), "num must be 0~9"

    def might_fail_func() -> pathlib.Path:
        return path_fusion(root_path, str(get_dataset()), num)

    if promise:
        p = func.try_again(
            15,
            might_fail_func,
            is_validate_func=lambda p: p.exists(),
            fail_message=f"cannot num: {num}",
        )
    else:
        p = might_fail_func()

    logger.debug(f"path: {p}")

    def yield_pics():
        gen = files.scan_pics(path=p)
        try:
            img = next(gen)
            yield img
        except StopIteration:
            raise Exception(f"no images found in dataset {p}")

    return yield_pics



import typing as t
import platform
import pathlib
import distro


def get_root_path():
    def get_system_type() -> t.Literal["macOS", "Windows", "Other", "Linux"]:
        """
        判断当前系统类型是 macOS 还是 Windows。

        Returns:
            str: 'macOS', 'Windows' 或 'Other'，分别表示 macOS 系统、Windows 系统或其他系统。
        """
        system = platform.system()

        if system == "Darwin":
            return "macOS"
        elif system == "Windows":
            return "Windows"
        elif system == "Linux":
            return "Linux"
        else:
            return "Other"

    if get_system_type() == "macOS":
        # expand userhome
        return pathlib.Path("/Volumes/Fast SSD/Research/Dataset/MeterData")
    elif get_system_type() == "Windows":
        return pathlib.Path(r"D:\Store\MeterData")
    elif get_system_type() == "Linux":
        if distro.version() == "24.04":
            return pathlib.Path("~/Work/Dataset/MeterData").expanduser()
        else:
            return pathlib.Path("~/work/Dataset/MeterData").expanduser()
    else:
        raise Exception("Not work under Linux now. 2024-05-21.")



import typing as t
from meterviewer import files
from meterviewer.datasets import dataset
import pathlib
from .types import viewReturn


def from_filesystem(root_path: pathlib.Path) -> viewReturn:
    for dataset_name in dataset.get_dataset_list(root_path):
        img_p = files.scan_pics(dataset.get_dataset_path(root_path, str(dataset_name)))
        img_p = list(img_p)
        for p in img_p:
            yield p, dataset_name, "invalid"


def view_dataset(root_path: pathlib.Path) -> viewReturn:
    def get_images(dataset_path: pathlib.Path):
        # lookup 3 images.
        pics = list(files.scan_pics(dataset_path))[:1]
        return pics

    datasets = dataset.get_dataset_list(root_path)
    for dataset_name in datasets:
        res = dataset.get_dataset_path(root_path, str(dataset_name))
        pics = get_images(res)
        for pic in pics:
            yield pic, dataset_name, "_"



import numpy as np
from meterviewer.img import validator


def test_validator():
    v = validator.Validator()

    cases = [
        (np.array([1, 2, 3]), False),
    ]
    for img, expected in cases:
        assert expected == v.is_img(img)



from meterviewer import types as T, files
from pathlib import Path as P

from meterviewer.datasets import dataset
from . import current


def more_quick_view(
    current_dataset: P,
    write_config=True,
):
    name_func = files.use_smart_name(current_dataset)
    return quick_view(str(current_dataset), name_func, write_config)


def quick_view(
    current_dataset: str,
    get_x_y_name: T.NameFunc,
    write_config=True,
):
    x_name, y_name = get_x_y_name()
    x = files.load_from_disk(P(current_dataset) / x_name)
    y = files.load_from_disk(P(current_dataset) / y_name)

    current.view_merge_np(current_dataset, get_x_y=get_x_y_name)
    if write_config:
        current.write_details(current_dataset, get_xy_name=get_x_y_name)
    return x, y


# from meterviewer.views.quick import fast_preview
def fast_preview(current_dataset: P):
    return more_quick_view(current_dataset)



# 统一数据集表示
# 只在一个文件里修改目录，方便快捷
# 将使用的函数放在参数上，减少依赖

from __future__ import annotations

import datetime
import os
import pathlib

# import numpy as np
import random
import typing as t

from tqdm import tqdm

from .. import files
from .. import types as T
from ..img import process
from .join import join_with_fix, join_with_resize  # noqa
from .view import view_dataset, view_dataset_in_rows, view_dataset_on_disk  # noqa


def generate_func() -> t.List[t.Callable[[pathlib.Path], T.ImgDataset]]:
    funcs = []
    names: t.List[str] = [T.x_name, T.y_name, T.x_test, T.y_test]
    for n in names:

        def get_func(path: pathlib.Path, name: str = n):
            return files.load_from_disk(path / name)

        funcs.append(get_func)

    return funcs


def get_details(
    path: pathlib.Path,
    x: T.ImgDataset,
    y: T.LabelData,
) -> t.Dict:
    data = {}

    def create_sub(name):
        data[name] = {}

    for name in ["Dataset", "Meta"]:
        create_sub(name)

    data["Dataset"]["path"] = str(path)
    data["Meta"]["config.create_time"] = datetime.datetime.now()
    data["Dataset"]["created_time"] = ""
    data["Dataset"]["updated_time"] = ""
    data["Dataset"]["x_shape"] = x.shape
    data["Dataset"]["y_shape"] = y.shape
    return data


def show_details(
    get_x_train: t.Callable,
    get_y_train: t.Callable,
    get_details: t.Callable[[t.Any, t.Any], t.Dict],
    write_to_file: t.Callable[[t.Dict], None],
):
    x, y = get_x_train(), get_y_train()
    details = get_details(x, y)
    write_to_file(details)


def dataset_length_list() -> t.List[int]:
    return [5, 6, 7, 8]


def create_str_with_fill(
    number: int,
    length: int,
    total: int,
) -> T.DigitStr:
    return fill_digit(process.number_to_string(number, length), total)


def create_labels_func(
    length: int,
    total: int,
) -> t.Callable[[int], t.Tuple[t.List[int], t.List[T.DigitStr]]]:
    """total length"""

    def generate_nums(train_nums: int):
        numbers = []
        str_digits = []
        for _ in range(train_nums):
            number = random.randint(0, 10**length)
            numbers.append(number)
            str_digits.append(
                create_str_with_fill(number, length, total),
            )
        return numbers, str_digits

    return generate_nums


class GenBlockImgFunc(t.Protocol):
    def __call__(self, digit: T.DigitStr) -> T.NpImage: ...


class SaveDatasetFunc(t.Protocol):
    def __call__(self, imgs: T.ImgList, str_digits: t.List[T.DigitStr]) -> None: ...


class CreateDatasetFunc(t.Protocol):
    def __call__(
        self,
        length: int,
        nums: int,
        gen_block_img: GenBlockImgFunc,
    ) -> t.Tuple[t.List[T.NpImage], t.List[T.DigitStr]]: ...


def create_dataset_func(
    check_imgs: t.Callable[[T.ImgList], None],
    total: int,
) -> CreateDatasetFunc:
    """创建新的数据库"""

    def inner(
        length: int,
        nums: int,
        gen_block_img: GenBlockImgFunc,
    ):
        _, str_digits = create_labels_func(length, total)(nums)

        imgs = []
        for digit in tqdm(str_digits):
            im = gen_block_img(digit)
            imgs.append(im)

        # files.write_shape(imgs, 3)
        check_imgs(imgs)

        # automatic resize the images
        imgs = process.resize_imglist(imgs, size=[37, 297])
        return imgs, str_digits

    return inner


def get_random_dataset(
    root: pathlib.Path, get_dataset_list: t.Callable
) -> t.Tuple[pathlib.Path, int]:
    datasets = list(get_dataset_list(root))
    random_index = random.randint(0, len(datasets) - 1)
    return datasets[random_index], random_index


def generate_block_img(
    the_digit: T.DigitStr,
    join_func: T.JoinFunc,
    read_rand_img: t.Callable[[int | str], T.NpImage],
) -> T.NpImage:
    img_list = []
    for digit in the_digit:
        im = read_rand_img(digit)
        img_list.append(im)
    return join_func(img_list, process.size_check)


def get_dataset_path(root: pathlib.Path, dataset_name: str) -> pathlib.Path:
    p = root / "lens_6" / "XL" / "XL" / dataset_name
    return p


def get_dataset_list(
    root: pathlib.Path,
    default_func: t.Callable = lambda: pathlib.Path("lens_6/XL/XL"),
) -> t.Iterator[pathlib.Path]:
    root = root / default_func()
    for dir in os.listdir(root):
        if os.path.isdir(root / dir):
            yield pathlib.Path(dir)


def handle_datasets(root: pathlib.Path, handle_func: t.Callable[[pathlib.Path], None]):
    """handle"""
    for dataset in get_dataset_list(root):
        handle_func(dataset)


def fill_digit(digit: T.DigitStr, total_length: int) -> T.DigitStr:
    digit2 = digit.copy()
    if len(digit) < 5:
        raise ValueError(f"digit lenght must > 5, {digit}")

    if len(digit) <= total_length:
        digit2.extend(["x"] * (total_length - len(digit)))
    else:
        raise ValueError(f"digit lenght must < {total_length}, {digit}")
    return digit2



import pytest
import pathlib
import typing as t
from pathlib import Path as P
from meterviewer.config import get_root_path


@pytest.fixture
def alldata() -> P:
    return P("alldata.db")


@pytest.fixture
def remove_files() -> t.Generator[t.Callable[[P], None], None, None]:
    filepath_list = []

    def func(filep: P):
        filepath_list.append(filep)

    yield func
    for filepath in filepath_list:
        filepath.unlink()


@pytest.fixture
def root_path() -> pathlib.Path:
    return get_root_path()